///////////////////////////////////////////////////////////////
//                                                           //
//  layer.cpp  -   Implementation file for the layer class   //
//  Language:      Visual C++, ver 5.0                       //
//  Platform:      Micron Dual Pentium 200, Windows NT 4.0   //
//  Author:        M. A. Saleh Elmohamed                     //
//                                                           //
///////////////////////////////////////////////////////////////
#include "layer.h"

layer::layer()
{
}

/*----< Size Constructor >-----------------------
  PROCESSING: Constructs a layer(output or normal) 
  with given no. of inputs and neurons.

  INPUTS:
  numInputs - Number of inputs the layer will have.
  numNeurons - Number of neurons the layer will have.
  outputLayer - Will this be an output layer ?

  OUTPUTS:
*/
layer::layer(int numInputs, int numNeurons, bool outputLayer)
{
  mi_numInputs = numInputs;
  mi_numNeurons = numNeurons;
  mb_outputLayer = outputLayer;

  vector<double> temp(mi_numNeurons);
  mvd_output = temp;

  neuron temp_neuron(numInputs, mb_outputLayer);

  for(int i=0; i<numNeurons; i++)
    mvn_neurons.push_back(temp_neuron);
}

layer::~layer()
{
}


//---- The next two operators are required by the STL vector class ----//

//------------ Comparison operator '==' ---------------
bool layer::operator==(const layer& l) const
{
  return false;
}

//------------ Comparison operator '<' ----------------
bool layer::operator<(const layer& l) const
{
  return true;
}

//------------ Assignment operator '=' ----------------
layer& layer::operator=(const layer& l)
{
  mi_numInputs = l.mi_numInputs;
  mi_numNeurons = l.mi_numNeurons;
  mvd_output = l.mvd_output;
  mvn_neurons = l.mvn_neurons;
  mb_outputLayer = l.mb_outputLayer;
}


/*----< Function to set the inputs of the layer >----
  PROCESSING: Sets the inputs of the layer to the 
  values in the supplied vectors.

  INPUTS:
  input_vec - The vector containing the values to which the inputs
  are to be set.

  OUTPUTS:
*/
void layer::set_inputs(vector<double> input_vec)
{
  if(input_vec.size()!=mi_numInputs)
    {
      cerr << "Number of inputs given does not match number of layer inputs";
      return;
    }
  for(int i=0; i<mi_numInputs; i++)
    for(int j=0; j<mi_numNeurons; j++)
      mvn_neurons[j].input(i) = input_vec[i];
}

/*----< Function to set the weights of the neurons in the layer >-----
  PROCESSING: Sets the weights of the layer to the values in the 
  supplied vectors.

  INPUTS:
  input_vec - The vector containing the values to which the weights
  are to be set.

  OUTPUTS:
*/
//---------- set the weights of the neurons in the layer ------
void layer::set_weights(vector< vector<double> > weight_vec)
{
  if(weight_vec.size()!=mi_numNeurons)
    {
      cerr << "Set of Weights dont match the number of neurons ";
      return;
    }
  for(int i=0; i<mi_numNeurons; i++)
    for(int j=0; j<mi_numInputs; j++)
      mvn_neurons[i].weight(j) = weight_vec[i][j];
}


//---------- Returns the weight vector of the layer -------
vector< vector<double> > layer::get_weights()
{
  vector< vector<double> > weight_vec(mi_numNeurons);
  vector<double> temp(mi_numInputs);

  for(int i=0; i<mi_numNeurons; i++)
    {
      weight_vec[i] = temp;
      for(int j=0; j<mi_numInputs; j++)
	weight_vec[i][j] = mvn_neurons[i].weight(j);
    }
  return weight_vec;
}


//---------- Sets the bias of the neurons in the layer ------
void layer::set_bias(vector<double> bias_vec)
{
  if(bias_vec.size()!=mi_numNeurons)
    {
      cerr << "Set of bias dont match the number of neurons ";
      return;
    }
  for(int i=0; i<mi_numNeurons; i++)
    mvn_neurons[i].bias() = bias_vec[i];
}


//---------- Returns the bias vector of the layer -----
vector<double> layer::get_bias()
{
  vector<double> bias_vec(mi_numNeurons);
  for(int i=0; i<mi_numNeurons; i++)
    bias_vec[i] = mvn_neurons[i].bias();

  return bias_vec;
}


//---------- Returns the o/p vector of the layer ------
vector<double> layer::get_outputs()
{
  vector<double> output_vec(mi_numNeurons);
  for(int i=0; i<mi_numNeurons; i++)
    output_vec[i] = mvn_neurons[i].find_output();

  return output_vec;
}


//----Returns a vector of propogation terms of the neurons in the layer ----
//----(The Sigma term in Equation 9.7 (see the report))
vector<double> layer::get_propTermVec()
{
  double sum;
  vector<double> propTerm_vec(mi_numInputs);
  for(int i=0; i<mi_numInputs; i++)
    {
      sum = 0.0;
      for(int j=0; j<mi_numNeurons; j++)
	sum += (mvn_neurons[j].weight(i) * mvn_neurons[j].error_term());
      propTerm_vec[i] = sum;
    }
  return propTerm_vec;
}


//----Returns a vector of error terms of the neurons in the layer ------
vector<double> layer::get_error_terms(vector<double> propTerm_vec)
{
  vector<double> errTerms_vec(mi_numNeurons);
  for(int i=0; i<mi_numNeurons; i++)
    errTerms_vec[i] = mvn_neurons[i].find_error_term(propTerm_vec[i]);
  
  return errTerms_vec;
}


//---- Updates the weights of all neurons ----
void layer::adapt()
{
  for(int i=0; i<mi_numNeurons; i++)
    mvn_neurons[i].update_weights();
}


//--------- Test Stub ---------------
#ifdef TEST_LAYER
void main()
{
  int i,j;
  vector<double> input(3), output(2), desired(2);
  vector< vector<double> > weight_vec;
  layer hidden_layer(3, 3);
  layer output_layer(3, 2, true);

  input[0] = 0.1; input[1] = 0.2; input[2] = 0.3;
  desired[0] = 0.6; desired[1] = 0.3;

  cout << "\tDesired1 : " << desired[0] << "\t";
  cout << "Desired2 : " << desired[1] << endl;
  hidden_layer.set_inputs(input);
  output_layer.set_inputs(hidden_layer.get_outputs());
  output = output_layer.get_outputs();
  for(i=0; i<250; i++)	{
    output_layer.get_error_terms(desired);
    hidden_layer.get_error_terms(output_layer.get_propTermVec());
    output_layer.adapt();
    hidden_layer.adapt();
    hidden_layer.set_inputs(input);
    output_layer.set_inputs(hidden_layer.get_outputs());
    output = output_layer.get_outputs();

    cout << "Output[" << i << "] : " << output[0] << "\t" << output[1] << endl;
  }
  cout << endl << "Weights" << endl;
  cout << "-------" << endl << endl;
  cout << "Hidden Layer" << endl;
  cout << "------------" << endl;
  weight_vec = hidden_layer.get_weights();
  for(i=0; i<3; i++)	{
    cout << "Neuron[" << i << "] :-" << endl;
    for(j=0; j<3; j++)
      cout << "\tWeight[" << j << "] = " << weight_vec[i][j] << endl;
    cout << endl;
  }
  weight_vec = output_layer.get_weights();
  cout << "Output Layer" << endl;
  cout << "------------" << endl;
  for(i=0; i<2; i++)	{
    cout << "Neuron[" << i << "] :-" << endl;
    for(j=0; j<3; j++)
      cout << "\tWeight[" << j << "] = " << weight_vec[i][j] << endl;
    cout << endl;
  }
}
#endif
