///////////////////////////////////////////////////////////////
//                                                           //
//  network.cpp -  Implementation file for the network class //
//  Language:      Visual C++, ver 5.0                       //
//  Platform:      Micron Dual Pentium 200, Windows NT 4.0   //
//  Author:        M. A. Saleh Elmohamed                     //
//                                                           //
///////////////////////////////////////////////////////////////
#include "network.h"

//-------- Size Constructor ---------------
//---- Constructs network with supplied parameters----
network::network(int numInputs,
		 int numOutputs, 
		 int numHiddenLayers, 
		 int numHiddenLayerNodes) 
  : ml_outputLayer(numHiddenLayerNodes, numOutputs, true),
    mvl_hiddenLayer(numHiddenLayers)
{
  mi_numInputs = numInputs;
  mi_numOutputs = numOutputs;
  mi_numHiddenLayers = numHiddenLayers;
  mi_numHiddenLayerNodes = numHiddenLayerNodes;

  //The first hidden layer will have the same no. of inputs as
  //the network has

  layer temp_layer1(numInputs, numHiddenLayerNodes);
  mvl_hiddenLayer[0] = temp_layer1;

  //For the other hidden layers, no. of inputs depend on
  //the no. of nodes in the layer

  layer temp_layer2(numHiddenLayerNodes, numHiddenLayerNodes);
  for(int i=1; i<numHiddenLayers; i++)
    mvl_hiddenLayer[i] = temp_layer2;
}

//--------- Destructor ----------------
network::~network()
{
}


//--------- Trains the network for a given i/o pair -------
vector<double> network::train(vector<double> input_vec, vector<double> desired_output_vec)
{
  vector<double> output_vec = get_outputs(input_vec);

  //Back propogation
  //First adapt the o/p layer

  ml_outputLayer.get_error_terms(desired_output_vec);

  //This will be used to adapt the last hidden layer
  
  mvl_hiddenLayer[mi_numHiddenLayers-1].get_error_terms(ml_outputLayer.get_propTermVec());

  ml_outputLayer.adapt();

  //Adapt the hidden layers
  
  for(int i=mi_numHiddenLayers-2; i>=0; i--)
    {
      mvl_hiddenLayer[i].get_error_terms(mvl_hiddenLayer[i+1].get_propTermVec());
      mvl_hiddenLayer[i+1].adapt();
    }

  //Lastly - adapt the first hidden layer
	
  mvl_hiddenLayer[0].adapt();
  return(output_vec);
}


//--------- Get the network output -------
vector<double> network::get_outputs(vector<double> input_vec)
{
  vector<double> output_vec(mi_numOutputs);
  if(input_vec.size()!=mi_numInputs)
    {
      cerr << "Number of inputs given does not match number of network inputs" << endl;
      return output_vec;
    }

  mvl_hiddenLayer[0].set_inputs(input_vec);

  //Propogating the o/p forward thru the hidden layers
  
  for(int i = 1; i<mi_numHiddenLayers; i++)
    mvl_hiddenLayer[i].set_inputs(mvl_hiddenLayer[i-1].get_outputs());

  //Applying o/p of last hidden layer as i/p of the output layer
  
  ml_outputLayer.set_inputs(mvl_hiddenLayer[mi_numHiddenLayers-1].get_outputs());
  output_vec = ml_outputLayer.get_outputs();
  return(output_vec);
}


//----- Sets the weights of the layers in the network -----
void network::set_weights(vector< layer_weights > weight_vecs)
{
  if(weight_vecs.size()!=(mi_numHiddenLayers + 1))
    {
      cerr << "Number of inputs given does not match number of layers";
      return;
    }
  for(int i=0; i<mi_numHiddenLayers; i++)
    mvl_hiddenLayer[i].set_weights(weight_vecs[i]);
  ml_outputLayer.set_weights(weight_vecs[mi_numHiddenLayers]);
}


//---------- Gets the weights of the layers in the network -----
vector< layer_weights > network::get_weights()
{
  vector< layer_weights > weight_vecs(mi_numHiddenLayers+1);

  for(int i=0; i<mi_numHiddenLayers; i++)
    weight_vecs[i] = mvl_hiddenLayer[i].get_weights();
  weight_vecs[mi_numHiddenLayers] = ml_outputLayer.get_weights();

  return weight_vecs;
}


//---------- Sets the bias of the layers in the network -----
void network::set_bias(vector< vector<double> > bias_vecs)
{
  if(bias_vecs.size()!=(mi_numHiddenLayers + 1))
    {
      cerr << "Number of inputs given does not match number of layers";
      return;
    }
  for(int i=0; i<mi_numHiddenLayers; i++)
    mvl_hiddenLayer[i].set_bias(bias_vecs[i]);
  ml_outputLayer.set_bias(bias_vecs[mi_numHiddenLayers]);
}


//---------- Gets the bias of the layers in the network -----
vector< vector<double> > network::get_bias()
{
  vector< vector<double> > bias_vecs(mi_numHiddenLayers+1);

  for(int i=0; i<mi_numHiddenLayers; i++)
    bias_vecs[i] = mvl_hiddenLayer[i].get_bias();
  bias_vecs[mi_numHiddenLayers] = ml_outputLayer.get_bias();

  return bias_vecs;
}


//--------- Test Stub ---------------
#ifdef TEST_NETWORK
void main()
{
  int i;

  vector<double> input(3);
  vector<double> output(2);
  vector<double> desired(2);
  desired[0]=(0.6);
  desired[1]=(0.3);

  cout << "  Desired :  " << desired[0] << "\t,\t" << desired[1];
  cout << endl << endl;

  network test_network(3, 2, 1, 3);
  input[0] = 0.1; input[1] = 0.2; input[2] = 0.3;

  for(i=0; i<350; i++)
    {
      output=test_network.train(input, desired);

      cout << "Output[" << i << "] : " << output[0] << "\t,\t" << output[1];
      cout << endl;
    }
  test_network.set_weights(test_network.get_weights());
}
#endif
