///////////////////////////////////////////////////////////////
//                                                           //
//  neuron.cpp  -  Implementation file for the neuron class  //
//  Language:      Visual C++, ver 5.0                       //
//  Platform:	   Micron Dual Pentium 200, Windows NT 4.0   //
//  Author:        M. A. Saleh Elmohamed                     //
//                                                           //
///////////////////////////////////////////////////////////////

#include "neuron.h"

//------- Void Constructor -------------------------------
neuron::neuron()
{
}


/*----< Size Constructor >------------------------
  PROCESSING: Constructs a neuron(output node or normal) 
  with 'n' inputs, assigns each input to 0, assigns the 
  supplied values to the input weights and bias.

  INPUTS:
  n - Number of inputs the neuron will have.
  out_node - Whether the neuron will be an output node.
  initial_weight - The value to which all the weights will be 
  initialized.
  initial_bias - The value to which the bias will be initialized.

  OUTPUTS:
*/
neuron::neuron(int n,
	       bool out_node,
	       double initial_weight,
	       double initial_bias)
{
  mi_numInputs = n;

  for(int i=0; i<mi_numInputs; i++)
    {
      mvd_input.push_back(0);
      mvd_weight.push_back(initial_weight);
    }

  md_bias = initial_bias;

  mb_outputNode = out_node;
}


//------- Destructor --------------
neuron::~neuron()
{
}


//------ The next three operators are required by the STL vector class ----//

//------------ Comparison operator '==' -----------
bool neuron::operator==(const neuron& n) const
{
  return false;
}


//------------ Comparison operator '<' ------------
bool neuron::operator<(const neuron& n) const
{
  return true;
}


//------------ Assignment operator '=' ------------
neuron& neuron::operator=(const neuron& n)
{
  mi_numInputs = n.mi_numInputs;
  mvd_input = n.mvd_input;
  mvd_weight = n.mvd_weight;
  md_bias = n.md_bias;
  md_output = n.md_output;
  md_errorTerm = n.md_errorTerm;
  mb_outputNode = n.mb_outputNode;
}


/*----< Function to access an input of the neuron >---------
  PROCESSING: Returns a reference to an input of the neuron

  INPUTS:
  index - The index of the input to be returned

  OUTPUTS:
  A reference to the input identified by the index supplied
*/

double& neuron::input(int index)
{
  static double safe;
  if(index<0 || index>mi_numInputs)
    {
      cerr << "Out of bounds";
      return safe;
    }
  return mvd_input[index];
}


/*----< Function to access a weight of the neuron >--------
  PROCESSING: Returns a reference to a weight of the neuron

  INPUTS:
  index - The index of the weight to be returned

  OUTPUTS:
  A reference to the weight identified by the index supplied
*/

double& neuron::weight(int index)
{
  static double safe;
  if(index<0 || index>mi_numInputs)
    {
      cerr << "Out of bounds";
      return safe;
    }
  return mvd_weight[index];
}


/*----< Function to access the bias of the neuron >-------
  PROCESSING: Returns a reference to the bias of the neuron

  INPUTS:

  OUTPUTS:
  A reference to the neuron's bias. 
*/

double& neuron::bias()
{
  return md_bias;
}


/*----< Function to generate neuron output >---------
  PROCESSING: Calculates and returns the neuron output.

  INPUTS:

  OUTPUTS:
  The calculated output
*/

double neuron::find_output()
{
  double out=0;
  for(int i=0;i<mi_numInputs;i++)
    out += (mvd_input[i] * mvd_weight[i]);

  out += md_bias;
  out = 1/(1+exp((-1)*out));

  md_output = out;
  return md_output;
}


/*----< Function to return neuron output >-------
  PROCESSING: Returns the last calculated neuron output, without any 
  re-calculation.

  INPUTS:

  OUTPUTS:
  The last calculated output.
*/

double neuron::output()
{
  return md_output;
}


/*----< Function to generate the error term(Epsilon) >-------
  PROCESSING: Calculates and returns the error term

  INPUTS:
  propogation_term - If the node is an output node, this will be the 
  desired output of the node for the given input.
  If the node is not an output node, it will be the Sigma term
  in Equation 9.7 (see the report).

  OUTPUTS:
  The calculated error term
*/
double neuron::find_error_term(double propogation_term)
{
  if(mb_outputNode)
    md_errorTerm = -2 * (propogation_term-md_output) * md_output * 
      (1 - md_output);
  else
    md_errorTerm = propogation_term * md_output * (1 - md_output);

  return md_errorTerm;
}


/*----< Function to return the error term >-------
  PROCESSING: Returns the last calculated error term, without any 
  re-calculation.

  INPUTS:

  OUTPUTS:
  The last calculated error term.
*/
double neuron::error_term()
{
  return md_errorTerm;
}


/*----< Function to update the neuron weights & bias >--------
  PROCESSING: calculates and updates the weights and bias of the 
              neuron

  INPUTS:

  OUTPUTS:

*/
void neuron::update_weights()
{
  double delta = 0;
  for(int i=0; i<mi_numInputs; i++)
    {
      //Eqn. 9.8 (see report)
      delta = -1 * ETA * md_errorTerm * mvd_input[i]; 
      mvd_weight[i] += delta;
    }

  delta = -1 * ETA * md_errorTerm;
  md_bias += delta;
}


//--------- Test Stub ---------------
#ifdef TEST_NEURON
void main()
{
  int n = 3;
  int i;
  double expected = 0;

  neuron testNeuron(n, true);
  for(i=0; i<n; i++)
    {
      testNeuron.input(i) = (double)(i+1)/10;
      cout << "Input " << i << " : " << testNeuron.input(i) << endl;
      expected += testNeuron.input(i);
    }
  cout << "Expected Output : " << expected << endl << endl;

  cout << "Output :" << testNeuron.find_output() << endl;

  for(i=0; i<250; i++)
    {
      testNeuron.find_error_term(expected);
      testNeuron.update_weights();
      cout << "Output " << i << " :" << testNeuron.find_output() << endl;
    }
}
#endif
